@@ -260,10 +260,14 @@ AppMenuButton.prototype = {
         this._iconBox = new Shell.Slicer({ name: 'appMenuIcon' });
         this._iconBox.connect('style-changed',
                               Lang.bind(this, this._onIconBoxStyleChanged));
+        this._iconBox.connect('notify::allocation',
+                              Lang.bind(this, this._updateIconBoxClip));
         this._container.add_actor(this._iconBox);
         this._label = new TextShadower();
         this._container.add_actor(this._label.actor);
 
+        this._iconBottomClip = 0;
+
         this._quitMenu = new PopupMenu.PopupMenuItem('');
         this.menu.addMenuItem(this._quitMenu);
         this._quitMenu.connect('activate', Lang.bind(this, this._onQuit));
@@ -300,6 +304,10 @@ AppMenuButton.prototype = {
 
         this._visible = true;
         this.actor.show();
+
+        if (!this._targetIsCurrent)
+            return;
+
         Tweener.removeTweens(this.actor);
         Tweener.addTween(this.actor,
                          { opacity: 255,
@@ -312,6 +320,11 @@ AppMenuButton.prototype = {
             return;
 
         this._visible = false;
+        if (!this._targetIsCurrent) {
+            this.actor.hide();
+            return;
+        }
+
         Tweener.removeTweens(this.actor);
         Tweener.addTween(this.actor,
                          { opacity: 0,
@@ -325,11 +338,16 @@ AppMenuButton.prototype = {
 
     _onIconBoxStyleChanged: function() {
         let node = this._iconBox.get_theme_node();
-        let bottomClip = node.get_length('app-icon-bottom-clip');
-        if (bottomClip > 0)
+        this._iconBottomClip = node.get_length('app-icon-bottom-clip');
+        this._updateIconBoxClip();
+    },
+
+    _updateIconBoxClip: function() {
+        let allocation = this._iconBox.allocation;
+        if (this._iconBottomClip > 0)
             this._iconBox.set_clip(0, 0,
-                                   this._iconBox.width,
-                                   this._iconBox.height - bottomClip);
+                                   allocation.x2 - allocation.x1,
+                                   allocation.y2 - allocation.y1 - this._iconBottomClip);
         else
             this._iconBox.remove_clip();
     },
@@ -622,12 +640,17 @@ PanelCorner.prototype = {
  * This class manages the "hot corner" that can toggle switching to
  * overview.
  */
-function HotCorner() {
-    this._init();
+function HotCorner(button) {
+    this._init(button);
 }
 
 HotCorner.prototype = {
-    _init : function() {
+    _init : function(button) {
+        // This is the activities button associated with this hot corner,
+        // if this is on the primary monitor (or null with the corner is
+        // on a different monitor)
+        this._button = button;
+
         // We use this flag to mark the case where the user has entered the
         // hot corner and has not left both the hot corner and a surrounding
         // guard area (the "environs"). This avoids triggering the hot corner
@@ -654,6 +677,8 @@ HotCorner.prototype = {
 
         this._activationTime = 0;
 
+        this.actor.connect('enter-event',
+                           Lang.bind(this, this._onEnvironsEntered));
         this.actor.connect('leave-event',
                            Lang.bind(this, this._onEnvironsLeft));
         // Clicking on the hot corner environs should result in the same bahavior
@@ -730,6 +755,11 @@ HotCorner.prototype = {
         this._addRipple(0.35,  1.0,   0.0,   0.3,    1,    0.0);
     },
 
+    _onEnvironsEntered : function() {
+        if (this._button)
+            this._button.hover = true;
+    },
+
     _onCornerEntered : function() {
         if (!this._entered) {
             this._entered = true;
@@ -757,6 +787,9 @@ HotCorner.prototype = {
     },
 
     _onEnvironsLeft : function(actor, event) {
+        if (this._button)
+            this._button.hover = false;
+
         if (event.get_related() != this._corner)
             this._entered = false;
         return false;
