@@ -101,8 +101,13 @@ NMNetworkMenuItem.prototype = {
         accessPoints = sortAccessPoints(accessPoints);
         this.bestAP = accessPoints[0];
 
-        let ssid = this.bestAP.get_ssid();
-        title = title || NetworkManager.utils_ssid_to_utf8(ssid) || _("<unknown>");
+        if (!title) {
+            let ssid = this.bestAP.get_ssid();
+            if (ssid)
+                title = NetworkManager.utils_ssid_to_utf8(ssid);
+            if (!title)
+                title = _("<unknown>");
+        }
 
         this._label = new St.Label({ text: title });
         this.addActor(this._label);
@@ -132,15 +137,18 @@ NMNetworkMenuItem.prototype = {
         }
     },
 
-    _updated: function(ap, strength) {
-        if (strength > this.bestAP.strength)
+    _updated: function(ap) {
+        if (ap.strength > this.bestAP.strength)
             this.bestAP = ap;
 
         this._signalIcon.icon_name = this._getIcon();
     },
 
     _getIcon: function() {
-        return 'network-wireless-signal-' + signalToIcon(this.bestAP.strength);
+        if (this.bestAP.mode == NM80211Mode.ADHOC)
+            return 'network-workgroup';
+        else
+            return 'network-wireless-signal-' + signalToIcon(this.bestAP.strength);
     },
 
     updateAccessPoints: function(accessPoints) {
@@ -200,7 +208,7 @@ NMDeviceTitleMenuItem.prototype = {
     },
 
     setStatus: function(text) {
-        if (text) {
+        if (text != null) {
             this._statusLabel.text = text;
             this._statusBin.child = this._statusLabel;
             this.actor.reactive = false;
@@ -367,6 +375,8 @@ NMDevice.prototype = {
                     this.deactivate();
                 this.emit('enabled-changed');
             }));
+
+            this._updateStatusItem();
         }
         this.section = new PopupMenu.PopupMenuSection();
 
@@ -383,10 +393,19 @@ NMDevice.prototype = {
             GObject.Object.prototype.disconnect.call(this.device, this._stateChangedId);
             this._stateChangedId = 0;
         }
+        if (this._carrierChangedId) {
+            // see above for why this is needed
+            GObject.Object.prototype.disconnect.call(this.device, this._carrierChangedId);
+            this._carrierChangedId = 0;
+        }
+        if (this._firmwareChangedId) {
+            GObject.Object.prototype.disconnect.call(this.device, this._firmwareChangedId);
+            this._firmwareChangedId = 0;
+        }
 
         this._clearSection();
-        if (this.titleItem)
-            this.titleItem.destroy();
+        if (this.statusItem)
+            this.statusItem.destroy();
         this.section.destroy();
     },
 
@@ -492,12 +511,16 @@ NMDevice.prototype = {
 
     getStatusLabel: function() {
         switch(this.device.state) {
+        case NetworkManager.DeviceState.UNMANAGED:
         case NetworkManager.DeviceState.DISCONNECTED:
+        case NetworkManager.DeviceState.DEACTIVATING:
         case NetworkManager.DeviceState.ACTIVATED:
             return null;
         case NetworkManager.DeviceState.PREPARE:
         case NetworkManager.DeviceState.CONFIG:
         case NetworkManager.DeviceState.IP_CONFIG:
+        case NetworkManager.DeviceState.IP_CHECK:
+        case NetworkManager.DeviceState.SECONDARIES:
             return _("connecting...");
         case NetworkManager.DeviceState.NEED_AUTH:
             /* Translators: this is for network connections that require some kind of key or password */
@@ -632,6 +655,14 @@ NMDevice.prototype = {
             break;
         }
 
+        this._updateStatusItem();
+
+        this._clearSection();
+        this._createSection();
+        this.emit('state-changed');
+    },
+
+    _updateStatusItem: function() {
         if (this._carrierChangedId) {
             // see above for why this is needed
             GObject.Object.prototype.disconnect.call(this.device, this._carrierChangedId);
@@ -644,10 +675,6 @@ NMDevice.prototype = {
 
         this.statusItem.setStatus(this.getStatusLabel());
         this.statusItem.setToggleState(this.connected);
-
-        this._clearSection();
-        this._createSection();
-        this.emit('state-changed');
     },
 
     _substateChanged: function() {
@@ -823,12 +850,15 @@ NMDeviceModem.prototype = {
     },
 
     _createSection: function() {
+        if (!this._shouldShowConnectionList())
+            return;
+
         if (this.mobileDevice) {
             // If operator_name is null, just pass the empty string, as the item is hidden anyway
             this._operatorItem = new PopupMenu.PopupImageMenuItem(this.mobileDevice.operator_name || '',
                                                                   this._getSignalIcon(),
                                                                   { reactive: false });
-            if (this.mobileDevice.operator_name)
+            if (!this.mobileDevice.operator_name)
                 this._operatorItem.actor.hide();
             this.section.addMenuItem(this._operatorItem);
         }
@@ -836,7 +866,7 @@ NMDeviceModem.prototype = {
         NMDevice.prototype._createSection.call(this);
     },
 
-    clearSection: function() {
+    _clearSection: function() {
         this._operatorItem = null;
 
         NMDevice.prototype._clearSection.call(this);
@@ -913,7 +943,7 @@ NMDeviceBluetooth.prototype = {
     },
 
     _updateAutoConnectionName: function() {
-        this._autoConnectionName = this._makeConnectioName(this.device);
+        this._autoConnectionName = this._makeConnectionName(this.device);
 
         this._clearSection();
         this._createSection();
@@ -1198,6 +1228,11 @@ NMDeviceWireless.prototype = {
             if (apObj.item)
                 apObj.item.destroy();
             this._networks.splice(pos, 1);
+            if (this._overflowItem &&
+                this._overflowItem.menu.length == 0) {
+                this._overflowItem.destroy();
+                this._overflowItem = null;
+            }
         } else if (apObj.item)
             apObj.item.updateAccessPoints(apObj.accessPoints);
     },
@@ -1657,6 +1692,8 @@ NMApplet.prototype = {
                 managedDevices.forEach(function(dev) {
                     dev.statusItem.actor.show();
                 });
+                // remove status text from the section title item
+                item.updateForDevice(null);
             }
         }
     },
@@ -1706,13 +1743,13 @@ NMApplet.prototype = {
                                                                 { icon: icon });
                 this._source.notify(notification);
             })); */
-            wrapper._stateChangedId = wrapper.connect('state-changed', Lang.bind(this, function(dev) {
+            wrapper._deviceStateChangedId = wrapper.connect('state-changed', Lang.bind(this, function(dev) {
                 this._syncSectionTitle(dev.category);
             }));
             wrapper._destroyId = wrapper.connect('destroy', function(wrapper) {
                 //wrapper.disconnect(wrapper._networkLostId);
                 //wrapper.disconnect(wrapper._activationFailedId);
-                wrapper.disconnect(wrapper._stateChangedId);
+                wrapper.disconnect(wrapper._deviceStateChangedId);
             });
             let section = this._devices[wrapper.category].section;
             let devices = this._devices[wrapper.category].devices;
@@ -1753,8 +1790,10 @@ NMApplet.prototype = {
 
         for (let i = 0; i < closedConnections.length; i++) {
             let active = closedConnections[i];
-            if (active._primaryDevice)
+            if (active._primaryDevice) {
                 active._primaryDevice.setActiveConnection(null);
+                active._primaryDevice = null;
+            }
             if (active._notifyStateId) {
                 active.disconnect(active._notifyStateId);
                 active._notifyStateId = 0;
@@ -2047,7 +2086,7 @@ NMApplet.prototype = {
                     let ap = dev.device.active_access_point;
                     let mode = dev.device.mode;
                     if (!ap) {
-                        if (mode != NetworkManager['80211Mode'].ADHOC) {
+                        if (mode != NM80211Mode.ADHOC) {
                             log('An active wireless connection, in infrastructure mode, involves no access point?');
                             break;
                         }
