@@ -1,4 +1,22 @@
 /* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*- */
+/*
+ * st-drawing-area.c: A dynamically-sized Cairo drawing area
+ *
+ * Copyright 2009, 2010 Red Hat, Inc.
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU Lesser General Public License as
+ * published by the Free Software Foundation, either version 2.1 of
+ * the License, or (at your option) any later version.
+ *
+ * This program is distributed in the hope it will be useful, but WITHOUT ANY
+ * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+ * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
+ * more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public License
+ * along with this program. If not, see <http://www.gnu.org/licenses/>.
+ */
 
 /**
  * SECTION:st-drawing-area
@@ -30,8 +48,8 @@
 G_DEFINE_TYPE(StDrawingArea, st_drawing_area, ST_TYPE_WIDGET);
 
 struct _StDrawingAreaPrivate {
-  CoglHandle *texture;
-  CoglHandle *material;
+  CoglHandle texture;
+  CoglHandle material;
   cairo_t *context;
   guint needs_repaint : 1;
   guint in_repaint : 1;
@@ -115,6 +133,7 @@ st_drawing_area_paint (ClutterActor *self)
           surface = cairo_image_surface_create (CAIRO_FORMAT_ARGB32, width, height);
           priv->context = cairo_create (surface);
           priv->in_repaint = TRUE;
+          priv->needs_repaint = FALSE;
 
           g_signal_emit ((GObject*)area, st_drawing_area_signals[REPAINT], 0);
 
