@@ -30,6 +30,7 @@
 
 #define CACHE_PREFIX_GICON "gicon:"
 #define CACHE_PREFIX_URI "uri:"
+#define CACHE_PREFIX_URI_FOR_CAIRO "uri-for-cairo:"
 #define CACHE_PREFIX_THUMBNAIL_URI "thumbnail-uri:"
 #define CACHE_PREFIX_RAW_CHECKSUM "raw-checksum:"
 #define CACHE_PREFIX_COMPRESSED_CHECKSUM "compressed-checksum:"
@@ -142,7 +143,7 @@ st_texture_cache_init (StTextureCache *self)
                                                    g_free, cogl_handle_unref);
   self->priv->outstanding_requests = g_hash_table_new_full (g_str_hash, g_str_equal,
                                                             g_free, NULL);
-  self->priv->thumbnails = gnome_desktop_thumbnail_factory_new (GNOME_DESKTOP_THUMBNAIL_SIZE_NORMAL);
+  self->priv->thumbnails = gnome_desktop_thumbnail_factory_new (GNOME_DESKTOP_THUMBNAIL_SIZE_LARGE);
 }
 
 static void
@@ -437,6 +438,67 @@ out:
   return rotated_pixbuf;
 }
 
+static GdkPixbuf*
+decode_image (const char *val)
+{
+  int i;
+  GError *error = NULL;
+  GdkPixbuf *res = NULL;
+  struct {
+    const char *prefix;
+    const char *mime_type;
+  } formats[] = {
+    { "data:image/x-icon;base64,", "image/x-icon" },
+    { "data:image/png;base64,", "image/png" }
+  };
+
+  g_return_val_if_fail (val, NULL);
+
+  for (i = 0; i < G_N_ELEMENTS (formats); i++)
+    {
+      if (g_str_has_prefix (val, formats[i].prefix))
+        {
+          gsize len;
+          guchar *data = NULL;
+          char *unescaped;
+
+          unescaped = g_uri_unescape_string (val + strlen (formats[i].prefix), NULL);
+          if (unescaped)
+            {
+              data = g_base64_decode (unescaped, &len);
+              g_free (unescaped);
+            }
+
+          if (data)
+            {
+              GdkPixbufLoader *loader;
+
+              loader = gdk_pixbuf_loader_new_with_mime_type (formats[i].mime_type, &error);
+              if (loader &&
+                  gdk_pixbuf_loader_write (loader, data, len, &error) &&
+                  gdk_pixbuf_loader_close (loader, &error))
+                {
+                  res = gdk_pixbuf_loader_get_pixbuf (loader);
+                  g_object_ref (res);
+                }
+              g_object_unref (loader);
+              g_free (data);
+            }
+        }
+    }
+  if (!res)
+    {
+      if (error)
+        {
+          g_warning ("%s\n", error->message);
+          g_error_free (error);
+        }
+      else
+        g_warning ("incorrect data uri");
+    }
+  return res;
+}
+
 static GdkPixbuf *
 impl_load_pixbuf_file (const char     *uri,
                        int             available_width,
@@ -448,6 +510,9 @@ impl_load_pixbuf_file (const char     *uri,
   char *contents = NULL;
   gsize size;
 
+  if (g_str_has_prefix (uri, "data:"))
+    return decode_image (uri);
+
   file = g_file_new_for_uri (uri);
   if (g_file_load_contents (file, NULL, &contents, &size, NULL, error))
     {
@@ -712,6 +777,7 @@ typedef struct {
   char *key;
   char *uri;
   gboolean thumbnail;
+  gboolean enforced_square;
   char *mimetype;
   GtkRecentInfo *recent_info;
   char *checksum;
@@ -723,15 +789,68 @@ typedef struct {
 } AsyncTextureLoadData;
 
 static CoglHandle
-pixbuf_to_cogl_handle (GdkPixbuf *pixbuf)
+pixbuf_to_cogl_handle (GdkPixbuf *pixbuf,
+                       gboolean   add_padding)
 {
-  return cogl_texture_new_from_data (gdk_pixbuf_get_width (pixbuf),
-                                     gdk_pixbuf_get_height (pixbuf),
-                                     COGL_TEXTURE_NONE,
-                                     gdk_pixbuf_get_has_alpha (pixbuf) ? COGL_PIXEL_FORMAT_RGBA_8888 : COGL_PIXEL_FORMAT_RGB_888,
-                                     COGL_PIXEL_FORMAT_ANY,
-                                     gdk_pixbuf_get_rowstride (pixbuf),
-                                     gdk_pixbuf_get_pixels (pixbuf));
+  CoglHandle texture, offscreen;
+  CoglColor clear_color;
+  int width, height;
+  guint size;
+
+  width = gdk_pixbuf_get_width (pixbuf);
+  height = gdk_pixbuf_get_height (pixbuf);
+  size = MAX (width, height);
+
+  if (!add_padding || width == height)
+    return cogl_texture_new_from_data (width,
+                                       height,
+                                       COGL_TEXTURE_NONE,
+                                       gdk_pixbuf_get_has_alpha (pixbuf) ? COGL_PIXEL_FORMAT_RGBA_8888 : COGL_PIXEL_FORMAT_RGB_888,
+                                       COGL_PIXEL_FORMAT_ANY,
+                                       gdk_pixbuf_get_rowstride (pixbuf),
+                                       gdk_pixbuf_get_pixels (pixbuf));
+
+  texture = cogl_texture_new_with_size (size, size,
+                                        COGL_TEXTURE_NO_SLICING,
+                                        COGL_PIXEL_FORMAT_ANY);
+
+  offscreen = cogl_offscreen_new_to_texture (texture);
+  cogl_color_set_from_4ub (&clear_color, 0, 0, 0, 0);
+  cogl_push_framebuffer (offscreen);
+  cogl_clear (&clear_color, COGL_BUFFER_BIT_COLOR);
+  cogl_pop_framebuffer ();
+  cogl_handle_unref (offscreen);
+
+  cogl_texture_set_region (texture,
+                           0, 0,
+                           (size - width) / 2, (size - height) / 2,
+                           width, height,
+                           width, height,
+                           gdk_pixbuf_get_has_alpha (pixbuf) ? COGL_PIXEL_FORMAT_RGBA_8888 : COGL_PIXEL_FORMAT_RGB_888,
+                           gdk_pixbuf_get_rowstride (pixbuf),
+                           gdk_pixbuf_get_pixels (pixbuf));
+  return texture;
+}
+
+static cairo_surface_t *
+pixbuf_to_cairo_surface (GdkPixbuf *pixbuf)
+{
+  cairo_surface_t *dummy_surface;
+  cairo_pattern_t *pattern;
+  cairo_surface_t *surface;
+  cairo_t *cr;
+
+  dummy_surface = cairo_image_surface_create (CAIRO_FORMAT_ARGB32, 1, 1);
+
+  cr = cairo_create (dummy_surface);
+  gdk_cairo_set_source_pixbuf (cr, pixbuf, 0, 0);
+  pattern = cairo_get_source (cr);
+  cairo_pattern_get_surface (pattern, &surface);
+  cairo_surface_reference (surface);
+  cairo_destroy (cr);
+  cairo_surface_destroy (dummy_surface);
+
+  return surface;
 }
 
 static GdkPixbuf *
@@ -796,7 +915,7 @@ on_pixbuf_loaded (GObject      *source,
   if (pixbuf == NULL)
     goto out;
 
-  texdata = pixbuf_to_cogl_handle (pixbuf);
+  texdata = pixbuf_to_cogl_handle (pixbuf, data->enforced_square);
 
   g_object_unref (pixbuf);
 
@@ -870,7 +989,7 @@ st_texture_cache_reset_texture (StTextureCachePropertyBind *bind,
 
   if (pixbuf != NULL)
     {
-      texdata = pixbuf_to_cogl_handle (pixbuf);
+      texdata = pixbuf_to_cogl_handle (pixbuf, FALSE);
       g_object_unref (pixbuf);
 
       clutter_texture_set_cogl_texture (bind->texture, texdata);
@@ -1097,6 +1216,7 @@ load_gicon_with_colors (StTextureCache    *cache,
       request->icon = g_object_ref (icon);
       request->icon_info = info;
       request->width = request->height = size;
+      request->enforced_square = TRUE;
 
       load_icon_pixbuf_async (cache, icon, info, size, colors, NULL, on_pixbuf_loaded, request);
     }
@@ -1109,6 +1229,8 @@ load_gicon_with_colors (StTextureCache    *cache,
        g_slist_free (request->textures);
        g_free (request);
        g_hash_table_remove (cache->priv->outstanding_requests, key);
+       g_object_unref (texture);
+       texture = NULL;
     }
 
   g_free (key);
@@ -1130,7 +1252,7 @@ load_gicon_with_colors (StTextureCache    *cache,
  * This will load @icon as a full-color icon; if you want a symbolic
  * icon, you must use st_texture_cache_load_icon_name().
  *
- * Return Value: (transfer none): A new #ClutterActor for the icon
+ * Return Value: (transfer none): A new #ClutterActor for the icon, or %NULL if not found
  */
 ClutterActor *
 st_texture_cache_load_gicon (StTextureCache    *cache,
@@ -1159,7 +1281,7 @@ load_from_pixbuf (GdkPixbuf *pixbuf)
 
   clutter_actor_set_size (CLUTTER_ACTOR (texture), width, height);
 
-  texdata = pixbuf_to_cogl_handle (pixbuf);
+  texdata = pixbuf_to_cogl_handle (pixbuf, FALSE);
 
   set_texture_cogl_texture (texture, texdata);
 
@@ -1305,6 +1427,46 @@ st_texture_cache_load_sliced_image (StTextureCache    *cache,
  * icon you are loading, use %ST_ICON_FULLCOLOR.
  */
 
+/* generates names like g_themed_icon_new_with_default_fallbacks(),
+ * but *only* symbolic names
+ */
+static char **
+symbolic_names_for_icon (const char *name)
+{
+  char **parts, **names;
+  int i, numnames;
+
+  parts = g_strsplit (name, "-", -1);
+  numnames = g_strv_length (parts);
+  names = g_new (char *, numnames + 1);
+  for (i = 0; parts[i]; i++)
+    {
+      if (i == 0)
+        {
+          names[i] = g_strdup_printf ("%s-symbolic", parts[i]);
+        }
+      else
+        {
+          names[i] = g_strdup_printf ("%.*s-%s-symbolic",
+                                      (int) (strlen (names[i - 1]) - strlen ("-symbolic")),
+                                      names[i - 1], parts[i]);
+        }
+    }
+  names[i] = NULL;
+
+  g_strfreev (parts);
+
+  /* need to reverse here, because longest (most specific)
+     name has to come first */
+  for (i = 0; i < (numnames / 2); i++) {
+    char *tmp = names[i];
+    names[i] = names[numnames - i - 1];
+    names[numnames - i - 1] = tmp;
+  }
+
+  return names;
+}
+
 /**
  * st_texture_cache_load_icon_name:
  * @cache: The texture cache instance
@@ -1328,24 +1490,41 @@ st_texture_cache_load_icon_name (StTextureCache    *cache,
 {
   ClutterActor *texture;
   GIcon *themed;
-  char *symbolic;
+  char **names;
 
   g_return_val_if_fail (!(icon_type == ST_ICON_SYMBOLIC && theme_node == NULL), NULL);
 
   switch (icon_type)
     {
     case ST_ICON_APPLICATION:
+      themed = g_themed_icon_new (name);
+      texture = load_gicon_with_colors (cache, themed, size, NULL);
+      g_object_unref (themed);
+      if (texture == NULL)
+        {
+          themed = g_themed_icon_new ("application-x-executable");
+          texture = load_gicon_with_colors (cache, themed, size, NULL);
+          g_object_unref (themed);
+        }
+      return CLUTTER_ACTOR (texture);
+      break;
     case ST_ICON_DOCUMENT:
       themed = g_themed_icon_new (name);
       texture = load_gicon_with_colors (cache, themed, size, NULL);
       g_object_unref (themed);
+      if (texture == NULL)
+        {
+          themed = g_themed_icon_new ("x-office-document");
+          texture = load_gicon_with_colors (cache, themed, size, NULL);
+          g_object_unref (themed);
+        }
 
       return CLUTTER_ACTOR (texture);
       break;
     case ST_ICON_SYMBOLIC:
-      symbolic = g_strconcat (name, "-symbolic", NULL);
-      themed = g_themed_icon_new_with_default_fallbacks ((const gchar*)symbolic);
-      g_free (symbolic);
+      names = symbolic_names_for_icon (name);
+      themed = g_themed_icon_new_from_names (names, -1);
+      g_strfreev (names);
       texture = load_gicon_with_colors (cache, themed, size,
                                         st_theme_node_get_icon_colors (theme_node));
       g_object_unref (themed);
@@ -1356,6 +1535,12 @@ st_texture_cache_load_icon_name (StTextureCache    *cache,
       themed = g_themed_icon_new_with_default_fallbacks (name);
       texture = load_gicon_with_colors (cache, themed, size, NULL);
       g_object_unref (themed);
+      if (texture == NULL)
+        {
+          themed = g_themed_icon_new ("image-missing");
+          texture = load_gicon_with_colors (cache, themed, size, NULL);
+          g_object_unref (themed);
+        }
 
       return CLUTTER_ACTOR (texture);
       break;
@@ -1423,7 +1608,7 @@ st_texture_cache_load_uri_sync_to_cogl_texture (StTextureCache *cache,
       if (!pixbuf)
         goto out;
 
-      texdata = pixbuf_to_cogl_handle (pixbuf);
+      texdata = pixbuf_to_cogl_handle (pixbuf, FALSE);
       g_object_unref (pixbuf);
 
       if (policy == ST_TEXTURE_CACHE_POLICY_FOREVER)
@@ -1440,6 +1625,45 @@ out:
   return texdata;
 }
 
+static cairo_surface_t *
+st_texture_cache_load_uri_sync_to_cairo_surface (StTextureCache        *cache,
+                                                 StTextureCachePolicy   policy,
+                                                 const gchar           *uri,
+                                                 int                    available_width,
+                                                 int                    available_height,
+                                                 GError               **error)
+{
+  cairo_surface_t *surface;
+  GdkPixbuf *pixbuf;
+  char *key;
+
+  key = g_strconcat (CACHE_PREFIX_URI_FOR_CAIRO, uri, NULL);
+
+  surface = g_hash_table_lookup (cache->priv->keyed_cache, key);
+
+  if (surface == NULL)
+    {
+      pixbuf = impl_load_pixbuf_file (uri, available_width, available_height, error);
+      if (!pixbuf)
+        goto out;
+
+      surface = pixbuf_to_cairo_surface (pixbuf);
+      g_object_unref (pixbuf);
+
+      if (policy == ST_TEXTURE_CACHE_POLICY_FOREVER)
+        {
+          cairo_surface_reference (surface);
+          g_hash_table_insert (cache->priv->keyed_cache, g_strdup (key), surface);
+        }
+    }
+  else
+    cairo_surface_reference (surface);
+
+out:
+  g_free (key);
+  return surface;
+}
+
 /**
  * st_texture_cache_load_uri_sync:
  *
@@ -1518,6 +1742,43 @@ st_texture_cache_load_file_to_cogl_texture (StTextureCache *cache,
   return texture;
 }
 
+/**
+ * st_texture_cache_load_file_to_cairo_surface:
+ * @cache: A #StTextureCache
+ * @file_path: Path to a file in supported image format
+ *
+ * This function synchronously loads the given file path
+ * into a cairo surface.  On error, a warning is emitted
+ * and %NULL is returned.
+ *
+ * Returns: (transfer full): a new #cairo_surface_t *
+ */
+cairo_surface_t *
+st_texture_cache_load_file_to_cairo_surface (StTextureCache *cache,
+                                             const gchar    *file_path)
+{
+  cairo_surface_t *surface;
+  GFile *file;
+  char *uri;
+  GError *error = NULL;
+
+  file = g_file_new_for_path (file_path);
+  uri = g_file_get_uri (file);
+
+  surface = st_texture_cache_load_uri_sync_to_cairo_surface (cache, ST_TEXTURE_CACHE_POLICY_FOREVER,
+                                                             uri, -1, -1, &error);
+  g_object_unref (file);
+  g_free (uri);
+
+  if (surface == NULL)
+    {
+      g_warning ("Failed to load %s: %s", file_path, error->message);
+      g_clear_error (&error);
+      return NULL;
+    }
+  return surface;
+}
+
 /**
  * st_texture_cache_load_file_simple:
  * @cache: A #StTextureCache
@@ -1598,7 +1859,7 @@ st_texture_cache_load_from_data (StTextureCache    *cache,
           return NULL;
         }
 
-      texdata = pixbuf_to_cogl_handle (pixbuf);
+      texdata = pixbuf_to_cogl_handle (pixbuf, TRUE);
       g_object_unref (pixbuf);
 
       set_texture_cogl_texture (texture, texdata);
@@ -1615,7 +1876,7 @@ st_texture_cache_load_from_data (StTextureCache    *cache,
 /**
  * st_texture_cache_load_from_raw:
  * @cache: a #StTextureCache
- * @data: raw pixel data
+ * @data: (array length=len): raw pixel data
  * @len: the length of @data
  * @has_alpha: whether @data includes an alpha channel
  * @width: width in pixels of @data
@@ -1722,6 +1983,7 @@ st_texture_cache_load_thumbnail (StTextureCache    *cache,
       data->thumbnail = TRUE;
       data->width = size;
       data->height = size;
+      data->enforced_square = TRUE;
       data->textures = g_slist_prepend (data->textures, g_object_ref (texture));
       load_thumbnail_async (cache, uri, mimetype, size, NULL, on_pixbuf_loaded, data);
     }
@@ -1798,6 +2060,7 @@ st_texture_cache_load_recent_thumbnail (StTextureCache    *cache,
       data->recent_info = gtk_recent_info_ref (info);
       data->width = size;
       data->height = size;
+      data->enforced_square = TRUE;
       data->textures = g_slist_prepend (data->textures, g_object_ref (texture));
       load_recent_thumbnail_async (cache, info, size, NULL, on_pixbuf_loaded, data);
     }
